"""
Vector Store on Weavite.
"""

from __future__ import annotations

__all__ = ["Weaviate"]

import copy
from typing import Any

import weaviate
import weaviate.classes as wvc
from weaviate.classes.query import Filter

from x8.core import Context, NCall, Operation, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError
from x8.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    QueryFunctionName,
    QueryProcessor,
    Select,
)
from x8.storage._common import (
    CollectionResult,
    CollectionStatus,
    ItemProcessor,
    ParameterParser,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
    VectorIndexMetric,
    VectorIndexStructure,
)

from .._helper import get_collection_config, is_metadata_field, is_value_field
from .._models import (
    VectorCollectionConfig,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorSearchArgs,
    VectorValue,
)

DEFAULT_LIMIT = 100


class Weaviate(StoreProvider):
    type: str
    collection: str | None
    dimension: int | dict[str, int]
    content_properties_field: str | None
    url: str | None
    api_key: str | None
    http_host: str
    http_port: int
    http_secure: bool
    grpc_host: str
    grpc_port: int
    grpc_secure: bool
    nparams: dict[str, Any]

    _database_client: weaviate.WeaviateClient
    _init: bool

    _processor: ItemProcessor

    _collection_cache: dict[str, WeaviateCollection]

    def __init__(
        self,
        type: str = "wcs",
        collection: str | None = None,
        dimension: int | dict[str, int] = 4,
        content_properties_field: str | None = "_content",
        url: str | None = None,
        api_key: str | None = None,
        http_host: str = "localhost",
        http_port: int = 8080,
        http_secure: bool = False,
        grpc_host: str = "localhost",
        grpc_port: int = 50051,
        grpc_secure: bool = False,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            type:
                Weaviate connection type.
                Values are "wcs", "local", "embedded", "custom".
                Defaults to "wcs".
            collection:
                Collection name mapped to vector store collection.
            dimension:
                Vector dimension.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the dimension.
            content_properties_field:
                Properties field where content is stored.
                Weaviate doesn't support a separate content field.
                Defaults to "_content".
            url:
                Weaviate url.
            api_key:
                Weaviate api key.
            http_host:
                Http host.
            http_port:
                Http port.
            http_secure:
                Secure http option.
            grpc_host:
                Grpc host.
            grpc_port:
                Grpc port.
            grpc_secure:
                Secure grpc option.
            nparams:
                Native parameters for weaviate client.
        """
        self.type = type
        self.collection = collection
        self.dimension = dimension
        self.content_properties_field = content_properties_field
        self.url = url
        self.api_key = api_key
        self.http_host = http_host
        self.http_port = http_port
        self.http_secure = http_secure
        self.grpc_host = grpc_host
        self.grpc_port = grpc_port
        self.grpc_secure = grpc_secure
        self.nparams = nparams

        self._init = False
        self._processor = ItemProcessor()
        self._collection_cache = dict()

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        if (
            self.type == "wcs"
            and self.url is not None
            and self.api_key is not None
        ):
            self._database_client = weaviate.connect_to_wcs(
                cluster_url=self.url,
                auth_credentials=weaviate.auth.AuthApiKey(self.api_key),
                **self.nparams,
            )
        elif self.type == "local":
            self._database_client = weaviate.connect_to_local(
                **self.nparams,
            )
        elif self.type == "embedded":
            self._database_client = weaviate.connect_to_embedded(
                **self.nparams,
            )
        elif self.type == "custom":
            self._database_client = weaviate.connect_to_custom(
                http_host=self.http_host,
                http_port=self.http_port,
                http_secure=self.http_secure,
                grpc_host=self.grpc_host,
                grpc_port=self.grpc_port,
                grpc_secure=self.grpc_secure,
                **self.nparams,
            )
        else:
            raise BadRequestError("Cannot initalize client")
        self._init = True

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            or op_parser.op_equals(StoreOperation.TRANSACT)
            else op_parser.get_collection_name()
        )
        collection = (
            collection_name or self.collection or self.__component__.collection
        )
        return collection

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> WeaviateCollection | None:
        if op_parser.is_resource_op():
            return None
        collection_name = self._get_collection_name(op_parser)
        if collection_name is None:
            raise BadRequestError("Collection name must be specified")
        if collection_name in self._collection_cache:
            return self._collection_cache[collection_name]
        client = self._database_client.collections.get(collection_name)
        dimension = ParameterParser.get_collection_parameter(
            self.dimension, collection_name
        )
        col = WeaviateCollection(
            client,
            collection_name,
            dimension,
            self.content_properties_field,
        )
        self._collection_cache[collection_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            Validator.validate_transact(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._database_client),
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: WeaviateCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if collection is not None:
            converter = collection.converter
            client = collection.client
            helper = collection.helper
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict[Any, Any] | None = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.has_collection, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_id_as_str())
            call = NCall(client.query.fetch_object_by_id, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args = {
                "id": op_parser.get_id_as_str(),
                "value": op_parser.get_vector_value(),
                "metadata": op_parser.get_metadata(),
                "nargs": nargs,
            }
            call = NCall(
                helper.put,
                args,
                None,
            )
        # UPDATE value
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args = converter.convert_update_value(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_value(),
            )
            call = NCall(client.data.update, args, nargs)
        # UPDATE metadata
        elif op_parser.op_equals(StoreOperation.UPDATE_METADATA):
            args = converter.convert_update_metadata(
                id=op_parser.get_id_as_str(),
                metadata=op_parser.get_metadata(),
            )
            call = NCall(client.data.update, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args, func = converter.convert_delete(
                id=op_parser.get_id_as_str_or_none(),
                where=op_parser.get_where(),
            )
            if func == "delete_by_id":
                call = NCall(client.data.delete_by_id, args, nargs)
            elif func == "delete_many":
                call = NCall(client.data.delete_many, args, nargs)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                search=op_parser.get_search_as_function(),
                where=op_parser.get_where(),
                select=op_parser.get_select(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(client.query.near_vector, args, nargs)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.aggregate.over_all, args, nargs)
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            args, func = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "get":
                call = NCall(client.query.fetch_objects, args, nargs)
            elif func == "put":
                call = NCall(client.data.insert_many, args, nargs)
            elif func == "delete":
                call = NCall(client.data.delete_many, args, nargs)
            state = {"func": func}
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: WeaviateCollection | None,
    ) -> Any:
        if collection is not None:
            processor = collection.processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = nresult
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = nresult
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = [col.lower() for col in list(nresult.keys())]
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            result = nresult
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if nresult is None:
                raise NotFoundError
            result = self._convert_to_item(nresult, processor)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
            )
        # UPDATE value
        elif op_parser.op_equals(StoreOperation.UPDATE):
            result = VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
        # UPDATE metadata
        elif op_parser.op_equals(StoreOperation.UPDATE_METADATA):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
                metadata=op_parser.get_metadata(),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult.objects:
                items.append(self._convert_to_item(item, processor))
            items = QueryProcessor.order_items(
                items, op_parser.get_order_by(), processor.resolve_root_field
            )
            result = VectorList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult.total_count
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            result = []
            func = state["func"] if state is not None else None
            if func == "get":
                for item in nresult.objects:
                    result.append(self._convert_to_item(item, processor))
            elif func == "put":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(
                        VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
                    )
            elif func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result

    def _convert_to_item(
        self, nitem: Any, processor: ItemProcessor
    ) -> VectorItem:
        content = None
        metadata: dict | None = None
        vector = None
        key = VectorKey(id=str(nitem.uuid))
        if nitem.properties is not None:
            metadata = nitem.properties
            if (
                metadata is not None
                and self.content_properties_field is not None
                and self.content_properties_field in metadata
            ):
                content = metadata[self.content_properties_field]
                metadata.pop(self.content_properties_field)
                if len(metadata) == 0:
                    metadata = None
        if nitem.vector is not None and len(nitem.vector) > 0:
            vector = nitem.vector["default"]
        value = VectorValue(
            vector=vector,
            sparse_vector=None,
            content=content,
        )
        properties = None
        if (
            nitem.metadata is not None
            and hasattr(nitem.metadata, "distance")
            and nitem.metadata.distance is not None
        ):
            properties = VectorProperties(score=nitem.metadata.distance * -1)
        return VectorItem(
            key=key, value=value, metadata=metadata, properties=properties
        )


class ResourceHelper:
    database_client: weaviate.WeaviateClient

    def __init__(self, database_client: weaviate.WeaviateClient):
        self.database_client = database_client

    def create_collection(
        self,
        collection: str,
        config: VectorCollectionConfig | None,
        exists: bool | None,
        nargs: Any,
    ) -> CollectionResult:
        if self.database_client.collections.exists(collection):
            if exists is False:
                raise ConflictError
            return CollectionResult(status=CollectionStatus.EXISTS)

        metric_mapping = {
            VectorIndexMetric.DOT_PRODUCT: wvc.config.VectorDistances.DOT,
            VectorIndexMetric.COSINE: wvc.config.VectorDistances.COSINE,
            VectorIndexMetric.EUCLIDEAN: wvc.config.VectorDistances.L2_SQUARED,
            VectorIndexMetric.MANHATTAN: wvc.config.VectorDistances.MANHATTAN,
            VectorIndexMetric.HAMMING: wvc.config.VectorDistances.HAMMING,
        }
        vector_index = config.vector_index if config is not None else None
        if vector_index is not None:
            metric = metric_mapping[vector_index.metric]
            structure = vector_index.structure
            nconfig = (
                vector_index.nconfig
                if vector_index.nconfig is not None
                else {}
            )
        else:
            metric = wvc.config.VectorDistances.DOT
            structure = None
            nconfig = {}
        if structure is None or structure == VectorIndexStructure.HNSW:
            vector_index_config: Any = wvc.config.Configure.VectorIndex.hnsw(
                distance_metric=metric, **nconfig
            )
        elif structure == VectorIndexStructure.FLAT:
            vector_index_config = wvc.config.Configure.VectorIndex.flat(
                distance_metric=metric, **nconfig
            )
        args = {
            "name": collection,
            "vectorizer_config": wvc.config.Configure.Vectorizer.none(),
            "vector_index_config": vector_index_config,
        }
        try:
            NCall(
                self.database_client.collections.create,
                args,
                nargs,
                {weaviate.exceptions.UnexpectedStatusCodeError: None},
            ).invoke()
        except weaviate.exceptions.UnexpectedStatusCodeError as e:
            if e.status_code == 422:
                return CollectionResult(status=CollectionStatus.EXISTS)
            raise e
        return CollectionResult(status=CollectionStatus.CREATED)

    def drop_collection(
        self,
        collection: str,
        exists: bool | None,
        nargs: Any,
    ) -> CollectionResult:
        if not self.database_client.collections.exists(collection):
            if exists is True:
                raise NotFoundError
            return CollectionResult(status=CollectionStatus.NOT_EXISTS)

        args = {"name": collection}
        NCall(self.database_client.collections.delete, args, nargs).invoke()
        return CollectionResult(status=CollectionStatus.DROPPED)

    def list_collections(self, nargs) -> Any:
        return NCall(
            self.database_client.collections.list_all, None, nargs
        ).invoke()

    def has_collection(
        self,
        collection: str,
        nargs: Any,
    ):
        args = {"name": collection}
        return NCall(
            self.database_client.collections.exists, args, nargs
        ).invoke()

    def close(self, nargs: Any) -> Any:
        return NCall(self.database_client.close, None, nargs).invoke()


class ClientHelper:
    client: Any
    converter: OperationConverter

    def __init__(self, client: Any, converter: OperationConverter) -> None:
        self.client = client
        self.converter = converter

    def put(self, id: str, value: dict, metadata: dict, nargs: Any) -> Any:
        args = self.converter.convert_put(id, value, metadata)
        try:
            return NCall(self.client.data.insert, args, nargs).invoke()
        except weaviate.exceptions.UnexpectedStatusCodeError as e:
            if e.status_code == 422:
                return NCall(self.client.data.replace, args, nargs).invoke()
            raise e

    def delete(self, id: str, where: Expression, nargs: Any) -> Any:
        if id is None and where is None:
            raise BadRequestError("Id or where should be provided for delete")
        if id is not None and where is None:
            args = {"uuid": id}
            return NCall(self.client.data.delete_by_id, args, nargs).invoke()
        if where is not None and id is None:
            args = {"where": self.converter.convert_expr(where)}
            return NCall(self.client.data.delete_many, args, nargs).invoke()
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )


class OperationConverter:
    processor: ItemProcessor
    collection: str | None
    dimension: int
    content_properties_field: str | None

    def __init__(
        self,
        processor: ItemProcessor,
        collection: str | None,
        dimension: int,
        content_properties_field: str | None,
    ) -> None:
        self.processor = processor
        self.collection = collection
        self.dimension = dimension
        self.content_properties_field = content_properties_field

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        batch_get = False
        batch_put = False
        batch_delete = False
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.GET):
                batch_get = True
            elif op_parser.op_equals(StoreOperation.PUT):
                batch_put = True
            elif op_parser.op_equals(StoreOperation.DELETE):
                batch_delete = True
        if (
            (batch_get and batch_put)
            or (batch_put and batch_delete)
            or (batch_delete and batch_get)
        ):
            raise BadRequestError(
                "Batch operation supports only one operation type"
            )
        if batch_get:
            return self.convert_batch_get(op_parsers), "get"
        elif batch_put:
            return self.convert_batch_put(op_parsers), "put"
        elif batch_delete:
            return self.convert_batch_delete(op_parsers), "delete"
        raise BadRequestError("Batch operation not supported")

    def convert_batch_get(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids: list = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {
            "filters": Filter.by_id().contains_any(ids),
            "include_vector": True,
        }

    def convert_batch_put(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        objs = list()
        for op_parser in op_parsers:
            objs.append(
                wvc.data.DataObject(
                    **self.convert_put(
                        op_parser.get_id_as_str(),
                        op_parser.get_vector_value(),
                        op_parser.get_metadata(),
                    )
                )
            )
        return {"objects": objs}

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids: list = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {"where": Filter.by_id().contains_any(ids)}

    def convert_get(self, id: str) -> dict:
        return {"uuid": id, "include_vector": True}

    def convert_put(self, id: str, value: dict, metadata: dict | None) -> dict:
        args: dict = {"uuid": id}
        content = None
        md = None
        vector_value = VectorValue(**value)
        if vector_value.vector is not None:
            args["vector"] = vector_value.vector

        content = vector_value.content
        if metadata is None:
            if (
                content is not None
                and self.content_properties_field is not None
            ):
                md = {self.content_properties_field: content}
        else:
            md = copy.deepcopy(metadata)
            if (
                content is not None
                and self.content_properties_field is not None
            ):
                md[self.content_properties_field] = content
        if md is not None:
            args["properties"] = md
        return args

    def convert_update_value(self, id: str, value: dict) -> dict:
        args: dict = {"uuid": id}
        vector_value = VectorValue(**value)
        if vector_value.vector is not None:
            args["vector"] = vector_value.vector
        if (
            vector_value.content is not None
            and self.content_properties_field is not None
        ):
            args["properties"] = {
                self.content_properties_field: vector_value.content
            }
        return args

    def convert_update_metadata(self, id: str, metadata: dict | None) -> dict:
        return {
            "uuid": id,
            "properties": metadata,
        }

    def convert_delete(self, id: str | None, where: Expression) -> Any:
        if id is None and where is None:
            raise BadRequestError("Id or where should be provided for delete")
        if id is not None and where is None:
            return {"uuid": id}, "delete_by_id"
        if where is not None and id is None:
            return {"where": self.convert_expr(where)}, "delete_many"
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )

    def convert_query(
        self,
        search: Function | None,
        where: Expression,
        select: Select | None,
        order_by: OrderBy | None,
        limit: int | None,
        offset: int | None,
    ) -> dict:
        args: dict = dict()
        search_args = self.convert_search(search)
        args = args | search_args
        select_args = self.convert_select(select)
        if select_args is not None:
            args = args | select_args
        if where is not None:
            args["filters"] = self.convert_expr(where)
        args["limit"] = limit
        args["offset"] = offset
        return args

    def convert_select(self, select: Select | None) -> dict:
        include_value = False
        if select is None or len(select.terms) == 0:
            include_value = True
        else:
            for term in select.terms:
                if is_value_field(term.field):
                    include_value = True
                elif is_metadata_field(term.field):
                    pass
        return {
            "include_vector": include_value,
            "return_metadata": wvc.query.MetadataQuery(
                distance=True, score=True
            ),
        }

    def convert_search(self, search: Function | None) -> dict:
        if search is None:
            return {"near_vector": [0.0] * self.dimension}
        namespace = search.namespace
        name = search.name
        if namespace == FunctionNamespace.BUILTIN:
            if name == QueryFunctionName.VECTOR_SEARCH:
                args = VectorSearchArgs(**search.named_args)
                return {
                    "near_vector": args.vector,
                }

        raise BadRequestError("Search function not supported")

    def convert_expr(self, expr: Expression) -> Any:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return expr
        if isinstance(expr, Field):
            field = self.processor.resolve_metadata_field(expr.path)
            return field
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            op = expr.op
            if isinstance(expr.lexpr, Field):
                field = self.convert_expr(expr.lexpr)
                value = self.convert_expr(expr.rexpr)
            elif isinstance(expr.rexpr, Field):
                field = self.convert_expr(expr.rexpr)
                value = self.convert_expr(expr.lexpr)
                op = Comparison.reverse_op(op)

            if op == ComparisonOp.EQ:
                return Filter.by_property(field).equal(value)
            elif op == ComparisonOp.NEQ:
                return Filter.by_property(field).not_equal(value)
            elif op == ComparisonOp.GT:
                return Filter.by_property(field).greater_than(value)
            elif op == ComparisonOp.GTE:
                return Filter.by_property(field).greater_or_equal(value)
            elif op == ComparisonOp.LT:
                return Filter.by_property(field).less_than(value)
            elif op == ComparisonOp.LTE:
                return Filter.by_property(field).less_or_equal(value)
            elif op == ComparisonOp.BETWEEN:
                return Filter.by_property(field).greater_or_equal(
                    value[0]
                ) & Filter.by_property(field).less_or_equal(value[1])
            elif op == ComparisonOp.IN:
                x = Filter.by_property(field).equal(value[0])
                for i in range(0, len(value)):
                    x = x | Filter.by_property(field).equal(value[i])
                return x
            elif op == ComparisonOp.NIN:
                x = Filter.by_property(field).not_equal(value[0])
                for i in range(0, len(value)):
                    x = x & Filter.by_property(field).not_equal(value[i])
                return x
            elif op == ComparisonOp.LIKE:
                return Filter.by_property(field).like(value)
            else:
                raise BadRequestError(f"Comparison op {expr.op} not supported")
        if isinstance(expr, And):
            return self.convert_expr(expr.lexpr) & self.convert_expr(
                expr.rexpr
            )
        if isinstance(expr, Or):
            return self.convert_expr(expr.lexpr) | self.convert_expr(
                expr.rexpr
            )
        if isinstance(expr, Not):
            raise BadRequestError("Not expression is not supported")
        return str(expr)

    def convert_func(self, expr: Function) -> Any:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == QueryFunctionName.IS_DEFINED:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).is_none(False)
            if name == QueryFunctionName.IS_NOT_DEFINED:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).is_none(True)
            if name == QueryFunctionName.ARRAY_CONTAINS:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).contains_all([args[1]])
            if name == QueryFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).contains_any(args[1])
        raise BadRequestError(f"Function {name} not supported")

    def convert_count(self, where: Expression) -> dict | None:
        if where is None:
            return None
        args = {"filters": self.convert_expr(where)}
        return args


class WeaviateCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor
    helper: ClientHelper

    def __init__(
        self,
        client: Any,
        collection: str | None,
        dimension: int,
        content_properties_field: str | None,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.converter = OperationConverter(
            processor=self.processor,
            collection=collection,
            dimension=dimension,
            content_properties_field=content_properties_field,
        )
        self.helper = ClientHelper(self.client, self.converter)



"""
Azure kubernetes deployment.
"""

import os
import re
import time
from pathlib import Path
from typing import Any, Iterable, Literal

import yaml
from kubernetes import client, config  # type: ignore
from x8.compute._common._image_helper import get_images
from x8.compute._common._models import ImageMap
from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Context, Operation, Provider, Response, RunContext

from .._models import ManifestsType


class BaseKubernetes(Provider):
    kubeconfig: str | dict[str, Any] | None
    context: str | None
    manifests: ManifestsType
    overlays: ManifestsType
    namespace: str | None
    images: list[ImageMap] | None
    containerizer: Containerizer | None
    container_registry: ContainerRegistry | None

    def __init__(
        self,
        kubeconfig: str | dict[str, Any] | None = None,
        context: str | None = None,
        manifests: ManifestsType = None,
        overlays: ManifestsType = None,
        namespace: str | None = None,
        images: list[ImageMap] | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        **kwargs: Any,
    ):
        self.kubeconfig = kubeconfig
        self.context = context
        self.manifests = manifests
        self.overlays = overlays
        self.namespace = namespace
        self.images = images
        self.containerizer = containerizer
        self.container_registry = container_registry
        super().__init__(**kwargs)

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        if not operation:
            self.__setup__(context)
            result = self.apply(run_context=self._get_run_context(context))
            return Response(result=result)
        return super().__run__(operation=operation, context=context, **kwargs)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        if not operation:
            await self.__asetup__(context)
            result = self.apply(run_context=self._get_run_context(context))
            return Response(result=result)
        return await super().__arun__(
            operation=operation, context=context, **kwargs
        )

    def _apply_provider_settings(
        self,
        objects: list[dict[str, Any]],
        image_map: dict[str, str],
    ):
        # Apply provider-specific settings to the object
        pass

    def _get_provider_kubeconfig(self) -> str | dict[str, Any] | None:
        return None

    def apply(
        self,
        manifests: ManifestsType = None,
        overlays: ManifestsType = None,
        namespace: str | None = None,
        server_side: bool = False,
        force_conflicts: bool = False,
        field_manager: str | None = None,
        dry_run: Literal["client", "server"] | None = None,
        validate: bool | Literal["strict"] | None = True,
        prune: bool = False,
        selector: str | None = None,
        prune_all: bool = False,
        prune_allowlist: list[str] | None = None,
        wait: bool = True,
        timeout: str | None = None,
        images: list[ImageMap] | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[None]:
        # Resolve inputs and load manifests/overlays
        base_manifests_input = self._get_manifests(manifests)
        overlay_manifests_input = self._get_overlays(overlays)
        kubeconfig_input = (
            self._get_kubeconfig() or self._get_provider_kubeconfig()
        )
        context_input = self._get_context()
        namespace_input = self._get_namespace(namespace)
        images_input = self._get_images(images)

        objects = self._load_manifests_input(base_manifests_input)
        overlay_objects = self._load_manifests_input(overlay_manifests_input)

        if overlay_objects:
            objects = self._apply_overlays(objects, overlay_objects)

        if not objects:
            return Response(result=None)

        image_map = dict()
        if images_input:
            image_uris = get_images(
                images=images_input,
                containerizer=self._get_containerizer(),
                container_registry=self._get_container_registry(),
                run_context=run_context,
            )
            for i, image in enumerate(images_input):
                image_map[image.name] = image_uris[i]
            self._rewrite_container_images(objects, image_map)

        self._apply_provider_settings(objects, image_map)

        # Initialize Kubernetes API client
        k8s_client = self._init_k8s_client(
            kubeconfig=kubeconfig_input, context=context_input
        )

        # Normalize and apply objects
        dry_run_opt = self._map_dry_run(dry_run)
        field_validation = self._map_field_validation(validate)
        timeout_seconds = self._parse_timeout(timeout) if wait else None

        applied_keys: set[tuple[str, str | None, str]] = set()
        for obj in objects:
            if not isinstance(obj, dict):
                continue
            # apply default namespace if provided and object is namespaced
            if namespace_input and self._is_namespaced_kind(obj):
                obj.setdefault("metadata", {}).setdefault(
                    "namespace", namespace_input
                )

            key = self._build_object_key(obj)
            applied_keys.add(key)

            if server_side:
                self._server_side_apply(
                    k8s_client,
                    obj,
                    field_manager=field_manager,
                    force=force_conflicts,
                    dry_run=dry_run_opt,
                    field_validation=field_validation,
                    default_namespace=namespace_input,
                )
            else:
                self._create_or_patch(
                    k8s_client,
                    obj,
                    dry_run=dry_run_opt,
                    field_validation=field_validation,
                    default_namespace=namespace_input,
                )

        # Optionally wait for readiness
        if wait:
            self._wait_for_objects(
                k8s_client,
                objects,
                namespace=namespace_input,
                timeout_seconds=timeout_seconds or 300,
            )

        # Optional prune based on selector
        if prune and selector:
            self._prune(
                k8s_client,
                desired_keys=applied_keys,
                selector=selector,
                namespace=namespace_input,
                prune_all=prune_all,
                prune_allowlist=prune_allowlist,
                dry_run=dry_run_opt,
            )
        return Response(result=None)

    def _get_run_context(self, context: Context | None = None) -> RunContext:
        if context and context.data:
            return context.data.pop("__run__", RunContext())
        return RunContext()

    def _get_kubeconfig(self) -> str | dict[str, Any] | None:
        if self.kubeconfig:
            return self.kubeconfig
        if self.__component__ and self.__component__.kubeconfig:
            return self.__component__.kubeconfig
        return None

    def _get_context(self) -> str | None:
        if self.context:
            return self.context
        if self.__component__ and self.__component__.context:
            return self.__component__.context
        return None

    def _get_manifests(
        self,
        manifests: ManifestsType,
    ) -> ManifestsType:
        if manifests:
            return manifests
        if self.manifests:
            return self.manifests
        if self.__component__ and self.__component__.manifests:
            return self.__component__.manifests
        return None

    def _get_overlays(
        self,
        overlays: ManifestsType,
    ) -> ManifestsType:
        if overlays:
            return overlays
        if self.overlays:
            return self.overlays
        if self.__component__ and self.__component__.overlays:
            return self.__component__.overlays
        return None

    def _get_namespace(self, namespace: str | None) -> str:
        if namespace:
            return namespace
        if self.namespace:
            return self.namespace
        if self.__component__ and self.__component__.namespace:
            return self.__component__.namespace
        return "default"

    def _get_images(self, images: list[ImageMap] | None) -> list[ImageMap]:
        if images:
            return images
        if self.images:
            return self.images
        if self.__component__ and self.__component__.images:
            return self.__component__.images
        return []

    def _get_containerizer(self) -> Containerizer | None:
        if self.containerizer:
            return self.containerizer
        if self.__component__ and self.__component__.containerizer:
            return self.__component__.containerizer
        return Containerizer(__provider__="default")

    def _get_container_registry(self) -> ContainerRegistry | None:
        if self.container_registry:
            return self.container_registry
        if self.__component__ and self.__component__.container_registry:
            return self.__component__.container_registry
        return ContainerRegistry(__provider__="default")

    def _init_k8s_client(
        self, kubeconfig: str | dict[str, Any] | None, context: str | None
    ) -> Any:
        """
        Initialize Kubernetes ApiClient using kubeconfig path or dict.
        Falls back to default config if none provided.
        """
        if isinstance(kubeconfig, dict):
            # load from dict if available
            try:
                # load_kube_config_from_dict is available in kubernetes>=18
                config.load_kube_config_from_dict(
                    config_dict=kubeconfig, context=context
                )
            except Exception:
                # fallback: may not support context arg
                config.load_kube_config_from_dict(kubeconfig)
        elif isinstance(kubeconfig, str) and kubeconfig:
            config.load_kube_config(config_file=kubeconfig, context=context)
        else:
            # default kubeconfig (e.g., ~/.kube/config) or in-cluster
            try:
                config.load_kube_config(context=context)
            except Exception:
                config.load_incluster_config()

        return client.ApiClient()

    def _load_manifests_input(
        self, manifests: ManifestsType
    ) -> list[dict[str, Any]]:
        """
        Load manifests from str path (file or dir), list of paths,
        dict, or list of dicts.
        """

        if manifests is None:
            return []

        def load_file(path: str) -> list[dict[str, Any]]:
            objs: list[dict[str, Any]] = []
            with open(path, "r", encoding="utf-8") as f:
                for doc in yaml.safe_load_all(f):
                    if isinstance(doc, dict):
                        objs.append(doc)
            return objs

        def load_dir(dir_path: str) -> list[dict[str, Any]]:
            objs: list[dict[str, Any]] = []
            for root, _, files in os.walk(dir_path):
                for name in sorted(files):
                    if name.endswith((".yml", ".yaml")):
                        objs.extend(load_file(os.path.join(root, name)))
            return objs

        if isinstance(manifests, str):
            p = Path(manifests)
            if p.is_dir():
                return load_dir(str(p))
            elif p.is_file():
                return load_file(str(p))
            else:
                # Not a valid path; treat as nothing
                return []
        if isinstance(manifests, list):
            if not manifests:
                return []
            if all(isinstance(x, str) for x in manifests):
                objs: list[dict[str, Any]] = []
                for s in manifests:
                    objs.extend(self._load_manifests_input(s))
                return objs
            if all(isinstance(x, dict) for x in manifests):
                return [x for x in manifests if isinstance(x, dict)]
            # Mixed or unsupported list
            return []
        if isinstance(manifests, dict):
            return [manifests]

        return []

    def _build_object_key(
        self, obj: dict[str, Any]
    ) -> tuple[str, str | None, str]:
        """Key to identify a K8s object: (gvk, namespace, name)."""
        api_version = obj.get("apiVersion", "")
        kind = obj.get("kind", "")
        meta = obj.get("metadata", {}) or {}
        name = meta.get("name", "")
        namespace = meta.get("namespace")
        return (f"{api_version}|{kind}", namespace, name)

    def _deep_merge(
        self, base: dict[str, Any], overlay: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Deep-merge overlay into base. Lists are replaced;
        dicts merged recursively.
        """
        for k, v in overlay.items():
            if v is None:
                # mimic kustomize behavior: explicit null removes field
                base.pop(k, None)
                continue
            if k not in base:
                base[k] = v
            else:
                if isinstance(base[k], dict) and isinstance(v, dict):
                    base[k] = self._deep_merge(base[k], v)
                else:
                    base[k] = v
        return base

    def _apply_overlays(
        self,
        base_objs: list[dict[str, Any]],
        overlay_objs: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        if not overlay_objs:
            return base_objs
        base_map: dict[tuple[str, str | None, str], dict[str, Any]] = {}
        for o in base_objs:
            if isinstance(o, dict):
                base_map[self._build_object_key(o)] = o
        for ovr in overlay_objs:
            key = self._build_object_key(ovr)
            if key in base_map:
                base_map[key] = self._deep_merge(base_map[key], ovr)
            else:
                base_map[key] = ovr
        # preserve original order as best-effort: base first, then new overlays
        merged: list[dict[str, Any]] = []
        seen: set[tuple[str, str | None, str]] = set()
        for obj in base_objs:
            k = self._build_object_key(obj)
            merged.append(base_map[k])
            seen.add(k)
        for k, v in base_map.items():
            if k not in seen:
                merged.append(v)
        return merged

    def _is_namespaced_kind(self, obj: dict[str, Any]) -> bool:
        # Heuristic: most core resources are namespaced except a few
        kind = (obj.get("kind") or "").lower()
        non_ns = {
            "namespace",
            "node",
            "persistentvolume",
            "clusterrole",
            "clusterrolebinding",
            "customresourcedefinition",
            "storageclass",
            "priorityclass",
            "mutatingwebhookconfiguration",
            "validatingwebhookconfiguration",
            "apiservice",
            "clusterissuer",
            "clusterpolicy",
        }
        return kind not in non_ns

    def _create_or_patch(
        self,
        api_client: Any,
        obj: dict[str, Any],
        dry_run: str | None,
        field_validation: str | None,
        default_namespace: str | None = None,
    ) -> None:
        # Determine API and method
        k8s_api, operation_kind, kwargs = self._resolve_api_and_kwargs(
            api_client, obj, default_namespace=default_namespace
        )

        # Try create, then patch if exists
        try:
            create_fn = getattr(k8s_api, f"create_{operation_kind}")
            create_kwargs: dict[str, Any] = {}
            if operation_kind.startswith("namespaced_"):
                create_kwargs["namespace"] = kwargs.get("namespace")
            if dry_run:
                create_kwargs["dry_run"] = dry_run
            if field_validation:
                create_kwargs["field_validation"] = field_validation
            create_fn(body=obj, **create_kwargs)
        except client.exceptions.ApiException as e:
            if e.status != 409:
                raise
            # Already exists -> patch
            patch_fn = getattr(k8s_api, f"patch_{operation_kind}")
            patch_kwargs = dict(kwargs)
            if dry_run:
                patch_kwargs["dry_run"] = dry_run
            if field_validation:
                patch_kwargs["field_validation"] = field_validation
            # name required for namespaced/global patch
            patch_fn(body=obj, **patch_kwargs)

    def _server_side_apply(
        self,
        api_client: Any,
        obj: dict[str, Any],
        field_manager: str | None,
        force: bool,
        dry_run: str | None,
        field_validation: str | None,
        default_namespace: str | None = None,
    ) -> None:
        k8s_api, operation_kind, kwargs = self._resolve_api_and_kwargs(
            api_client, obj, default_namespace=default_namespace
        )

        patch_fn = getattr(k8s_api, f"patch_{operation_kind}")
        params = dict(kwargs)
        if dry_run:
            params["dry_run"] = dry_run
        if field_manager:
            params["field_manager"] = field_manager
        if force:
            params["force"] = True
        if field_validation:
            params["field_validation"] = field_validation

        # Content type for Server-Side Apply
        params["_content_type"] = "application/apply-patch+yaml"
        patch_fn(body=obj, **params)

    def _resolve_api_and_kwargs(
        self,
        api_client: Any,
        obj: dict[str, Any],
        default_namespace: str | None = None,
    ) -> tuple[Any, str, dict[str, Any]]:
        """
        Return (api, operation_suffix, kwargs)
        where operation_suffix is like 'namespaced_deployment' or
        'deployment'. kwargs includes namespace/name for namespaced
        operations and name for cluster-scoped.
        """

        kind = obj.get("kind", "")
        group, _, version = (obj.get("apiVersion", "") or "").partition("/")
        if version == "":
            version = group
            group = "core"
        group = "".join(group.rsplit(".k8s.io", 1))
        group = "".join(word.capitalize() for word in group.split("."))
        api_name = f"{group}{version.capitalize()}Api"
        k8s_api = getattr(client, api_name)(api_client)

        # snake_case kind

        def to_snake(s: str) -> str:
            s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
            s = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s)
            return s.lower()

        skind = to_snake(kind)
        metadata = obj.get("metadata", {}) or {}
        name = metadata.get("name")
        namespace = metadata.get("namespace")

        namespaced_method = f"create_namespaced_{skind}"
        # methods used for probing existence
        # (not referenced directly to avoid flake warnings)

        kwargs: dict[str, Any] = {}
        # Determine if namespaced by probing for method existence
        if hasattr(k8s_api, namespaced_method):
            # namespaced
            kwargs["namespace"] = namespace or default_namespace or "default"
            if name:
                kwargs["name"] = name
            operation_kind = f"namespaced_{skind}"
        else:
            # cluster scoped
            if name:
                kwargs["name"] = name
            operation_kind = skind
        return k8s_api, operation_kind, kwargs

    def _map_dry_run(
        self, dry_run: Literal["client", "server"] | None
    ) -> str | None:
        # Kubernetes API uses dry_run="All" for both client/server intent;
        # it always happens server-side.
        return "All" if dry_run else None

    def _map_field_validation(
        self, validate: bool | Literal["strict"] | None
    ) -> str | None:
        if validate is True or validate == "strict":
            return "Strict"
        if validate is False or validate is None:
            return None
        return None

    def _parse_timeout(self, timeout: str | None) -> int:
        if not timeout:
            return 300
        # parse formats like 30s, 2m, 1h
        m = re.match(r"^(\d+)([smh]?)$", timeout.strip())
        if not m:
            return 300
        value = int(m.group(1))
        unit = m.group(2)
        if unit == "m":
            return value * 60
        if unit == "h":
            return value * 3600
        return value

    def _wait_for_objects(
        self,
        api_client: Any,
        objects: list[dict[str, Any]],
        namespace: str | None = None,
        timeout_seconds: int = 300,
    ) -> None:
        start = time.time()
        for obj in objects:
            kind = (obj.get("kind") or "").lower()
            meta = obj.get("metadata", {}) or {}
            name = meta.get("name")
            namespace = meta.get("namespace") or namespace

            if kind in {
                "deployment",
                "statefulset",
                "daemonset",
                "job",
                "cronjob",
            }:
                if isinstance(name, str) and name:
                    self._wait_kind(
                        api_client,
                        kind,
                        name,
                        namespace,
                        timeout_seconds,
                        start,
                    )

    def _wait_kind(
        self,
        api_client: Any,
        kind: str,
        name: str,
        namespace: str | None,
        timeout_seconds: int,
        start_time: float,
    ) -> None:
        def elapsed() -> float:
            return time.time() - start_time

        if kind == "deployment":
            api = client.AppsV1Api(api_client)
            while elapsed() < timeout_seconds:
                # type: ignore[arg-type]
                dep = api.read_namespaced_deployment(name, namespace)
                spec_repl = (dep.spec.replicas or 1) if dep and dep.spec else 1
                avail = (
                    dep.status.available_replicas
                    if (
                        dep
                        and dep.status
                        and dep.status.available_replicas is not None
                    )
                    else 0
                )
                if avail >= spec_repl:
                    return
                time.sleep(2)
        elif kind == "statefulset":
            api = client.AppsV1Api(api_client)
            while elapsed() < timeout_seconds:
                # type: ignore[arg-type]
                st = api.read_namespaced_stateful_set(name, namespace)
                spec_repl = (st.spec.replicas or 1) if st and st.spec else 1
                ready = (
                    st.status.ready_replicas
                    if (
                        st
                        and st.status
                        and st.status.ready_replicas is not None
                    )
                    else 0
                )
                if ready >= spec_repl:
                    return
                time.sleep(2)
        elif kind == "daemonset":
            api = client.AppsV1Api(api_client)
            while elapsed() < timeout_seconds:
                # type: ignore[arg-type]
                ds = api.read_namespaced_daemon_set(name, namespace)
                desired = (
                    ds.status.desired_number_scheduled
                    if (
                        ds
                        and ds.status
                        and ds.status.desired_number_scheduled is not None
                    )
                    else 0
                )
                ready = (
                    ds.status.number_ready
                    if ds and ds.status and ds.status.number_ready is not None
                    else 0
                )
                if desired > 0 and ready >= desired:
                    return
                time.sleep(2)
        elif kind == "job":
            batch = client.BatchV1Api(api_client)
            while elapsed() < timeout_seconds:
                jb = batch.read_namespaced_job(name, namespace)
                succeeded = (
                    jb.status.succeeded
                    if jb and jb.status and jb.status.succeeded is not None
                    else 0
                )
                if succeeded and succeeded > 0:
                    return
                time.sleep(2)
        elif kind == "cronjob":
            batch = client.BatchV1Api(api_client)
            while elapsed() < timeout_seconds:
                cj = batch.read_namespaced_cron_job(name, namespace)
                # Consider cronjob "ready" if at least one job
                # was scheduled and none active
                last_schedule = (
                    cj.status.last_schedule_time if cj and cj.status else None
                )
                active = (
                    len(cj.status.active)
                    if cj and cj.status and cj.status.active
                    else 0
                )
                if last_schedule and active == 0:
                    return
                time.sleep(2)

    def _prune(
        self,
        api_client: Any,
        desired_keys: set[tuple[str, str | None, str]],
        selector: str,
        namespace: str | None,
        prune_all: bool,
        prune_allowlist: list[str] | None,
        dry_run: str | None,
    ) -> None:
        # Build list of kinds to consider:
        # either from allowlist or from desired set
        kinds_in_desired = {k.split("|")[1] for (k, _, _) in desired_keys}
        consider_kinds = (
            set(prune_allowlist) if prune_allowlist else kinds_in_desired
        )

        # Map kind to list and delete APIs
        api_map: list[tuple[str, Any, Any]] = []
        core = client.CoreV1Api(api_client)
        apps = client.AppsV1Api(api_client)
        batch = client.BatchV1Api(api_client)
        # Only a subset for safety
        api_map.extend(
            [
                ("Pod", core.list_namespaced_pod, core.delete_namespaced_pod),
                (
                    "Service",
                    core.list_namespaced_service,
                    core.delete_namespaced_service,
                ),
                (
                    "ConfigMap",
                    core.list_namespaced_config_map,
                    core.delete_namespaced_config_map,
                ),
                (
                    "Secret",
                    core.list_namespaced_secret,
                    core.delete_namespaced_secret,
                ),
                (
                    "PersistentVolumeClaim",
                    core.list_namespaced_persistent_volume_claim,
                    core.delete_namespaced_persistent_volume_claim,
                ),
                (
                    "Deployment",
                    apps.list_namespaced_deployment,
                    apps.delete_namespaced_deployment,
                ),
                (
                    "StatefulSet",
                    apps.list_namespaced_stateful_set,
                    apps.delete_namespaced_stateful_set,
                ),
                (
                    "DaemonSet",
                    apps.list_namespaced_daemon_set,
                    apps.delete_namespaced_daemon_set,
                ),
                (
                    "Job",
                    batch.list_namespaced_job,
                    batch.delete_namespaced_job,
                ),
            ]
        )

        for kind, list_fn, delete_fn in api_map:
            if kind not in consider_kinds:
                continue
            if namespace is None:
                # Require namespace for namespaced prune to avoid broad deletes
                continue
            objs = list_fn(namespace=namespace, label_selector=selector)
            for item in objs.items or []:
                gvk = self._infer_gvk_from_obj(item)
                name = item.metadata.name
                key = (gvk, namespace, name)
                if key not in desired_keys:
                    params: dict[str, Any] = {}
                    if dry_run:
                        params["dry_run"] = dry_run
                    try:
                        delete_fn(name=name, namespace=namespace, **params)
                    except client.exceptions.ApiException as e:
                        if e.status == 404:
                            continue
                        else:
                            raise

    def _infer_gvk_from_obj(self, item: Any) -> str:
        api_version = getattr(item, "api_version", None) or getattr(
            item, "apiVersion", ""
        )
        kind = getattr(item, "kind", "")
        return f"{api_version}|{kind}"

    # --- Helpers to rewrite image uris ---
    def _iter_podspecs(self, obj: dict[str, Any]) -> Iterable[dict[str, Any]]:
        """Yield every PodSpec dict contained in a single K8s object."""
        kind = (obj.get("kind") or "").lower()
        spec = obj.get("spec") or {}

        if kind == "pod":
            if spec:
                yield spec
            return

        if kind in {
            "deployment",
            "replicaset",
            "replicationcontroller",
            "statefulset",
            "daemonset",
            "job",
        }:
            tmpl = spec.get("template") or {}
            ps = tmpl.get("spec") or {}
            if ps:
                yield ps
            return

        if kind == "cronjob":
            jt = spec.get("jobTemplate") or {}
            jts = jt.get("spec") or {}
            tmpl = jts.get("template") or {}
            ps = tmpl.get("spec") or {}
            if ps:
                yield ps
            return
        # Other kinds: ignore

    def _rewrite_container_images_on_obj(
        self, obj: dict[str, Any], images_map: dict[str, str]
    ) -> None:
        """Rewrite images on a single object in place."""
        for ps in self._iter_podspecs(obj):
            for section in (
                "containers",
                "initContainers",
                "ephemeralContainers",
            ):
                arr = ps.get(section)
                if not isinstance(arr, list):
                    continue
                for c in arr:
                    if not isinstance(c, dict):
                        continue
                    cname = c.get("name")
                    cur = c.get("image")
                    if not isinstance(cname, str) or not isinstance(cur, str):
                        continue
                    if cname in images_map:
                        new_img = images_map[cname]
                        c["image"] = new_img
                        c["imagePullPolicy"] = "IfNotPresent"

    def _rewrite_container_images(
        self, objects: list[dict[str, Any]], images_map: dict[str, str]
    ) -> None:
        """
        Rewrite images across a list of objects.
        """
        for obj in objects:
            if (
                isinstance(obj, dict)
                and obj.get("kind") == "List"
                and isinstance(obj.get("items"), list)
            ):
                for item in obj["items"]:
                    if isinstance(item, dict):
                        self._rewrite_container_images_on_obj(item, images_map)
            elif isinstance(obj, dict):
                self._rewrite_container_images_on_obj(obj, images_map)



from typing import Any

from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Context, Operation, Provider, Response, RunContext
from x8.core.exceptions import BadRequestError
from x8.ql import Expression

from .._helper import merge_service_overlay
from .._models import ServiceDefinition, ServiceItem, ServiceOverlay


class BaseContainerDeploymentProvider(Provider):
    service: ServiceDefinition | None
    overlay: ServiceOverlay | None
    containerizer: Containerizer | None
    container_registry: ContainerRegistry | None

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        if not operation:
            self.__setup__(context)
            result = self.create_service(
                run_context=self._get_run_context(context)
            )
            return Response(result=result)
        return super().__run__(operation=operation, context=context, **kwargs)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        if not operation:
            await self.__asetup__(context)
            result = await self.acreate_service(
                run_context=self._get_run_context(context)
            )
            return Response(result=result)
        return await super().__arun__(
            operation=operation, context=context, **kwargs
        )

    def create_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        raise NotImplementedError(
            "Create service method must be implemented by provider."
        )

    async def acreate_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        raise NotImplementedError(
            "Create service method must be implemented by provider."
        )

    def _get_run_context(self, context: Context | None = None) -> RunContext:
        if context and context.data:
            return context.data.pop("__run__", RunContext())
        return RunContext()

    def _get_containerizer(self) -> Containerizer:
        return (
            self.containerizer
            or self.__component__.containerizer
            or Containerizer(__provider__="default")
        )

    def _get_container_registry(self) -> ContainerRegistry | None:
        return (
            self.container_registry
            or self.__component__.container_registry
            or None
        )

    def _deploy(
        self,
        service: ServiceDefinition,
        images: list[str],
        where_exists: bool | None,
    ) -> ServiceItem:
        raise NotImplementedError(
            "Deploy method must be implemented by provider."
        )

    def _normalize_service_definition(
        self, service: ServiceDefinition | None = None
    ) -> ServiceDefinition:
        service_def = service or self.service or self.__component__.service
        if not service_def:
            raise BadRequestError("Service definition is required.")
        service_def = merge_service_overlay(
            service=service_def, overlay=self.overlay
        )
        return service_def
